////////////////////////////////////////////////////////////////////////////////////////////////////////
// This code is a part of NOAA STAR Ocean Color MSL12 destriping package.                             //
// Developed by Karlis Mikelsons at NOAA/NESDIS/STAR and GST, INC.                 12/12/2014         //
// Please see the README file for credits and scknowledgments.                                        //
////////////////////////////////////////////////////////////////////////////////////////////////////////

#ifndef READWRITE_MSL12_HDF4
#define READWRITE_MSL12_HDF4

#include <stdio.h>
#include <string.h>
#include "mfhdf.h"
#define MAX_STR_LEN 1024

////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////
template <class dtype>
int read_msl12_hdf4(dtype ** buffer, int * nx, int * ny, char * sds_name, char * filename, int readwrite = 0) {

    ////////////////////////////////////////////////////////////////////////////////////////////////////
    // this finction reads or writes (parameter readwrite = 0 or 1, respectively)
    // a two-dimensional array of data 
    // identified by string sds_name in hdf4 file
    // from/to hdf file with name given by string filename
    //
    // on return, nx and ny are the dimensions of the two-dimensional array read/written
    // with ny being the major index and nx the minor index
    // so that entry (ix,iy) corresponds to location buffer[0][iy*nx+ix]
    // for (0 <= ix < nx) and (0 <= iy < ny)
    //
    // in case of read:
    //   this function allocates buffer where data are read, and points buffer[0] to this buffer;
    //   is it up to caller to deallocate this buffer later
    //
    // in case of write:
    //   buffer[0] points to data that should be written when this function is called;
    //   the buffer of data remains intact
    //
    // on success, the return value is 0;
    // nonzero value means problem of some kind
    ////////////////////////////////////////////////////////////////////////////////////////////////////

    int32 sd_id, sds_index, sds_id; /* SD interface and data set identifiers */
    intn status; /* status returned by some routines; has value SUCCEED or FAIL */
    int i, j, k, iprint = 0;  // set iprint > 0 for debug printout

    if(iprint > 0) printf("read_msl12_hdf4\n");     

    // open SD interface for read / write access
    if(readwrite==0) { sd_id = SDstart(filename, DFACC_READ); } 
    else             { sd_id = SDstart(filename, DFACC_WRITE); }
    if(sd_id==FAIL) {  
        // cannot open file - bad
        if(iprint > 0) printf("Cannot open %s with SDstart\n", filename);     
        return -1;  
    }

    // find the index of field given by name in sds_name
    sds_index = SDnametoindex(sd_id, sds_name);
    if(sds_index==FAIL) {  

        // cannot find the field
        if(iprint > 0) printf("Cannot get index of %s with SDnametoindex\n", sds_name);  

        // close the file, so it does not get locked up
        status = SDend(sd_id);
        if(status==FAIL) {     
            // cannot even close it !?
            if(iprint > 0) printf("Cannot end with SDend\n");     
            return -1;
        }
        // return 1 - means everything is still OK, just could not find a dataset
        return 1;  
    }

    // select the field, given the index
    sds_id = SDselect(sd_id, sds_index);
    if(sds_id==FAIL) {  
        if(iprint > 0) printf("Cannot select data set with SDselect\n");
        status = SDend(sd_id);
        if(status==FAIL) {     
            // cannot even close it !?
            if(iprint > 0) printf("Cannot end with SDend\n");     
        }
        return -1;  
    }

    // some work variables
    char sds_name1[MAX_STR_LEN];
    int32 rank = 0, data_type = 0, num_attrs = 0;
    int32 dimsizes[32];  // at most 32 dimensions, but that shouls not happen
    for(i=0; i<32; i++) dimsizes[i] = 0;

    // get the info about the field - such as dimensions size etc
    status = SDgetinfo(sds_id, sds_name1, &rank, dimsizes, &data_type, &num_attrs);
    if(status==FAIL) { 
        if(iprint > 0) printf("Cannot get info about %s with SDgetinfo\n", sds_name1);
        return -1;  
    }

    if(iprint > 0) printf("rank = %i\n", rank);

    // make sure we have two-dimensional array
    if(rank!=2) { 
        // we anticipated two dimensions, but got something else
        printf("ERROR: %s rank = %i != 2\n", sds_name, rank);
        return -1; 
    }
    
    if(iprint > 0) for(i=0; i<rank; i++) printf("%i %i\n",i,dimsizes[i]);
    if(iprint > 0) printf("datatype = %i\n", data_type);

    // set the return values for the dimension sizes
    *ny = dimsizes[0];
    *nx = dimsizes[1];
    if(iprint>0) printf("nx = %i ny = %i\n", *nx, *ny);
  
    // set the data needed for read / write 
    int32 start[2]  = { 0, 0 };                          // always start from (0,0) element
    int32 stride[2] = { 1, 1 };                          // do not skip any rows or columns
    int32 edge[2]   = { dimsizes[0], dimsizes[1] };      // set the size equal to full array size
    int ntot = dimsizes[0] * dimsizes[1];                // get the size of array to allocate (in case of read)

    if(readwrite==0) {

        // reading data:
        // allocate the buffer for data
        buffer[0] = NULL;
        buffer[0] = (dtype *) malloc(ntot*sizeof(dtype));
        if(buffer[0]==NULL) { 
            // cannot allocate data - error
            if(iprint > 0) printf("Cannot allocate memory %i bytes\n", ntot*sizeof(dtype)); 
            return -1; 
        }

        // read data
        status = SDreaddata(sds_id, start, stride, edge, (VOIDP) buffer[0]);
        if (status==FAIL) {
            if(iprint > 0) printf("Cannot   read data with SDreaddata \n");
            return -1; 
        }
    }
    else {
 
        // write data
        status = SDwritedata(sds_id, start, stride, edge, (VOIDP) buffer[0]);
        if(status==FAIL) {
            // could not write data
            if(iprint > 0) printf("Cannot  write data with SDwritedata %i %i %i %i %i %i\n", 
                                  start[0], start[1], stride[0], stride[1], edge[0], edge[1]);
            return -1;
        }
    }

    // end access to file
    status = SDendaccess(sds_id);
    if(status==FAIL) {
        if(iprint > 0) printf("Cannot end access with SDendaccess\n");
        return -1;
    }

    // end using SD interface
    status = SDend(sd_id);
    if(status==FAIL) {
        if(iprint > 0) printf("Cannot end with SDend\n");
        return -1;
    }

    return 0;

};



////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////
template <class dtype>
int       write_msl12_hdf4(dtype ** buffer, int * nx, int * ny, char * sds_name, char * filename) {
    return read_msl12_hdf4(         buffer,       nx,       ny,        sds_name,        filename, /* readwrite = */ 1 );
};


////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////
int write_msl12_hdf4_attr_strarr(char * filename, char * attr_name, int nstr, char ** str_arr){

    if(nstr<=0) return 0;

    int i, iprint = 0;   /* set iprint > 0 for debug printout */
    intn status;      /* status returned by some routines; has value SUCCEED or FAIL */

    // oen hdf4 file
    int32 sd_id = SDstart(filename, DFACC_WRITE);
    if(sd_id==FAIL) {  if(iprint > 0) printf("Cannot open %s with SDstart\n", filename);     return 1;  }

    // try to see if the attribute exists
    intn attr_index = SDfindattr (sd_id, attr_name);

    char attrbuff[MAX_STR_LEN];
    for (i=0; i<MAX_STR_LEN; i++) { attrbuff[i] = 0; }

    if(attr_index!=FAIL) {
        if (iprint > 0) printf("Attribute %s     found in file %s\n", attr_name, filename);

        // we need to read the values of this attribute and modify those of destriped bands
        status = SDreadattr (sd_id, attr_index, attrbuff);
        if(status==FAIL) { 
            // attribute found, but cannot read it - problem
            printf("Cannot read attr %s with SDreadattr\n", attr_name);  
            return -1; 
        }
        else             { 
            // print warning message for attributes already destriped
            printf("WARNING: Following products were already destriped: %s = %s\n", attr_name, attrbuff); 
        }
    }
    else {
        // did not find attribute in the hdf file - that's OK, will add it
        if(iprint > 0) printf("Attribute %s not found in file %s\n", attr_name, filename);
    }

    // modify attribute for destriped products
    // concatenate strings in array str_arr
    int istr, nchar = 0;
    attrbuff[0] = '\0';
    for(istr=0; istr<nstr; istr++){
        if(istr>0) strcat(attrbuff," \0");
        strcat(attrbuff, str_arr[istr]);
    }
    nchar = strlen(attrbuff);

    // write the new attribute to the hdf file
    if(iprint > 0) printf("Writing destriping attribute %s = %s\n", attr_name, attrbuff);
    status = SDsetattr(sd_id, attr_name, DFNT_CHAR, nchar, attrbuff);
    if(status==FAIL) {
        if(iprint > 0) printf("Cannot write attribute with SDsetattr\n");
        return -1;  
    }

    // end access to file
    status = SDend(sd_id);
    if(status==FAIL) {
        if(iprint > 0) printf("Cannot end access with SDend\n");
        return -1; 
    }

    return 0;
};

#endif
